import json
import click
from kube_compose import utils
from kube_compose.cli import cli

@cli.command(context_settings={"ignore_unknown_options": True})
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.argument('args', nargs=-1, type=str, metavar='[SERVICE:SRC_PATH|SRC_PATH|-] [DEST_PATH|-|SERVICE:DEST_PATH]')
def cp(*, args, namespace, kubectl, **_):
  ''' Like `docker-compose cp` but for the kubernetes deployed resources
  '''
  args_ = []
  for arg in args:
    if ':' in arg:
      service, _, path = arg.partition(':')
      # find the pod for the specified docker-compose service
      pod = utils.check_output([
        *kubectl, 'get', 'pod',
        *(['-n', namespace] if namespace else []),
        '-l', f"app.kubernetes.io/name={service}", '-o', 'jsonpath="{.items[0][\'metadata.name\']}"'
      ])
      args_.append(
        (f"{namespace}/" if namespace else '')
        + f"{json.loads(pod.decode())}:{path}"
      )
    else:
      args_.append(arg)
  utils.run([*kubectl, 'cp', *args_])



import click
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.argument('service', type=str, required=True)
@click.argument('args', nargs=-1, type=str)
def restart(service, args, *, namespace, kubectl, **_):
  ''' port-forward service [local_port:]remote_port ...
  '''
  utils.run([
    *kubectl, 'port-forward',
    *(('-n', namespace) if namespace else tuple()),
    f"deploy/{service}",
    *args,
  ])



import click
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.option('-p', '--previous', type=bool, is_flag=True, help='Print the logs for the previous instance of the container')
@click.option('-f', '--follow', type=bool, is_flag=True, help='Follow log output')
@click.option('-n', '--tail', type=int, default=-1, help='Number of lines to show from the end of the logs (-1 all lines)')
@click.option('-t', '--timestamps', type=bool, is_flag=True, help='Show timestamps')
@click.argument('service', type=str)
@click.argument('args', nargs=-1, type=str)
def logs(service, args, *, previous, follow, tail, timestamps, kubectl, namespace, **_):
  ''' Like `docker-compose logs` but for the kubernetes deployed resources
  '''
  utils.run([
    *kubectl, 'logs',
    *(('-p',) if previous else tuple()),
    *(('-f',) if follow else tuple()),
    *((f"--tail={tail}",) if tail != -1 else tuple()),
    *((f"--timestamps",) if timestamps else tuple()),
    *(('-n', namespace) if namespace else tuple()),
    f"deploy/{service}",
    *args,
  ])



import click
import itertools
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.argument('args', nargs=-1, type=str)
def scale(args, *, namespace, kubectl, **_):
  ''' Like `docker-compose scale` but effects the kubernetes deployed resources
  '''
  service_replicas = []
  for arg in args:
    svc, _, replicas = arg.partition('=')
    service_replicas.append((int(replicas), svc))
  for replicas, svcs in itertools.groupby(sorted(service_replicas), lambda key: key[0]):
    utils.run([
      *kubectl, 'scale',
      *(('-n', namespace) if namespace else tuple()),
      f"--replicas={replicas}",
      *[f"deploy/{svc}" for svc in svcs],
    ])



import click
from kube_compose import utils
from kube_compose.cli.volume import volume

@volume.command('inspect')
@click.argument('volume', type=str, required=True)
def _(**kwargs):
  ''' Remove a kubernetes persistent volume
  '''
  inspect(**kwargs)

@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
def inspect(*, volume, namespace, kubectl, **_):
  utils.run([
    *kubectl, 'get',
    *(('-n', namespace) if namespace else tuple()),
    f"pvc/{volume}",
    '-o', 'json',
  ])



import sys
import click
import yaml
from kube_compose import utils
from kube_compose.cli.volume import volume

def create_volume_spec(*, volume, docker_compose_config, **_):
  volume_specs = []
  for volume, volume_config in ([(volume, docker_compose_config['volumes'][volume],)] if volume is not None else docker_compose_config.get('volumes', {}).items()):
    volume_ext_config = volume_config.get('x-kubernetes', {})
    volume_specs.append({
      'apiVersion': 'v1',
      'kind': 'PersistentVolumeClaim',
      'metadata': {
        'name': volume,
      },
      'spec': {
        'accessModes': [volume_ext_config.get('mode', 'ReadWriteOnce')],
        'resources': {
          'requests': {
            'storage': volume_ext_config.get('size', '1Gi'),
          },
        },
        'storageClassName': volume_ext_config.get('class', ''),
        'volumeMode': 'Filesystem',
      },
    })
  return volume_specs

@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
def create(*, volume, docker_compose_config, namespace, kubectl, **_):
  volume_specs = create_volume_spec(
    volume=volume,
    docker_compose_config=docker_compose_config,
    kubectl=kubectl,
  )
  if not volume_specs: return
  utils.run([
    *kubectl, 'apply',
    *(('-n', namespace) if namespace else tuple()),
    '-f', '-',
  ], input=yaml.dump_all(volume_specs).encode())

@volume.command('create')
@click.argument('volume', type=str, required=False)
def _(**kwargs):
  ''' Ensure a kubernetes persistent volume is created
  '''
  create(**kwargs)

@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
def template(*, volume, docker_compose_config, namespace, kubectl, **_):
  volume_specs = create_volume_spec(
    volume=volume,
    docker_compose_config=docker_compose_config,
    kubectl=kubectl,
  )
  if not volume_specs: return
  # use kubectl to fill-in defaults/validate
  templ = yaml.safe_load(utils.check_output([
    *kubectl, 'apply',
    '--dry-run=client',
    '-oyaml',
    *(('-n', namespace) if namespace else tuple()),
    '-f', '-',
  ], input=yaml.dump_all(create_volume_spec(
    volume=volume,
    docker_compose_config=docker_compose_config,
    kubectl=kubectl,
  )).encode()))
  # write items in flat yaml format
  sys.stdout.write('---\n')
  if templ['kind'] == 'List':
    yaml.dump_all(templ['items'], sys.stdout)
  else:
    yaml.dump(templ, sys.stdout)

@volume.command('template')
@click.argument('volume', type=str, required=False)
def _(**kwargs):
  ''' Ensure a kubernetes persistent volume is templated
  '''
  template(**kwargs)



import click
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.argument('service', type=str, required=False)
def stop(service, *, namespace, kubectl, docker_compose_config, **_):
  ''' Like `docker-compose stop` but effects the kubernetes deployed resources
  '''
  services = [service] if service is not None else docker_compose_config.get('services', {}).keys()
  utils.run([
    *kubectl, 'scale',
    *(('-n', namespace) if namespace else tuple()),
    '--replicas=0',
    *[f"deploy/{service}" for service in services],
  ])



import subprocess
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(helm='helm', kubectl='kubectl')
@utils.require_kube_compose_release
def up(*, helm, kubectl, name, namespace, docker_compose_config_raw, docker_compose_config, **_):
  ''' Like `docker-compose up` but runs in the kubernetes cluster
  '''
  from kube_compose.cli.volume.create import create
  create(volume=None)
  if subprocess.call([
    *helm, 'status', *(('-n', namespace) if namespace else tuple()), name
  ], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0:
    utils.run([
      *helm, 'upgrade',
      *(('-n', namespace) if namespace else tuple()),
      name,
      utils.helm_chart,
      '-f', '-'
    ], input=docker_compose_config_raw)
  else:
    utils.run([
      *helm, 'install',
      *(('-n', namespace) if namespace else tuple()),
      name,
      utils.helm_chart,
      '-f', '-',
    ], input=docker_compose_config_raw)
  #
  utils.run([
    *kubectl, 'rollout', 'status',
    *(('-n', namespace) if namespace else tuple()),
    *[f"deploy/{service}" for service in docker_compose_config.get('services', {})],
  ])



import click
from datetime import datetime
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.argument('service', type=str, required=False)
def top(service, *, namespace, docker_compose_config, kubectl, **_):
  ''' Like `docker-compose top` but show status of the kubernetes deployed resource
  '''
  import subprocess
  for svc in ([service] if service is not None else docker_compose_config.get('services', {}).keys()):
    print(svc)
    try:
      results = subprocess.check_output([
        *kubectl, 'exec', 
        *(('-n', namespace) if namespace else tuple()),
        f"deploy/{svc}",
        '--',
        '/bin/sh',
        '-c',
        'cat /proc/*/stat 2> /dev/null; exit 0'
      ])
    except subprocess.CalledProcessError as e:
      print('  Error checking processes')
    else:
      print(f"{'UID':>6}{'PID':>10}{'PPID':>10}{'C':>5}{'STIME':>8}{'TTY':>6}{'TIME':>11} {'CMD'}")
      for line in results.decode().splitlines():
        # https://man7.org/linux/man-pages/man5/proc.5.html
        (
          pid,
          comm,
          state,
          ppid,
          pgrp,
          session,
          tty_nr,
          tpgid,
          flags,
          minflt,
          cminflt,
          majflt,
          cmajflt,
          utime,
          stime,
          cutime,
          cstime,
          priority,
          nice,
          num_threads,
          itrealvalue,
          starttime,
          vsize,
          rss,
          rsslim,
          *_,
        ) = line.split(' ')
        print(f"{'?':>6}{pid:>10}{ppid:>10}{'?':>5}{starttime:>8}{'?':>6}{utime:>11} {comm[1:-1]}")



from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(helm='helm')
@utils.require_kube_compose_release
def template(*, helm, name, namespace, docker_compose_config_raw, **_):
  ''' Alias for `helm template [release]`
  '''
  from kube_compose.cli.volume.create import template as volume_create_template
  volume_create_template(volume=None)
  utils.run([
    *helm, 'template',
    *(('-n', namespace) if namespace else tuple()),
    name,
    utils.helm_chart,
    '-f', '-', '--validate=false',
  ], input=docker_compose_config_raw)



import click
from kube_compose.cli import cli
from kube_compose import utils

@cli.command(context_settings={"ignore_unknown_options": True})
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.option('-i', '--stdin', type=bool, is_flag=True)
@click.option('-t', '--tty', type=bool, is_flag=True)
@click.argument('service', type=str)
@click.argument('args', nargs=-1, type=str)
def exec(service, args, *, stdin, tty, namespace, kubectl, **_):
  ''' Like `docker-compose exec` but for the kubernetes deployed resources
  '''
  utils.run([
    *kubectl, 'exec',
    *(('-n', namespace) if namespace else tuple()),
    *(('-i',) if stdin else tuple()),
    *(('-t',) if tty else tuple()),
    f"deploy/{service}",
    *(('--', *args,) if args else tuple()),
  ])



import click
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.argument('service', type=str, required=False)
def stats(service, *, namespace, docker_compose_config, kubectl, **_):
  ''' Like `docker-compose stats` but show status of the kubernetes deployed resource
  '''
  services = [service] if service is not None else docker_compose_config.get('services', {}).keys()
  utils.run([
    *kubectl, 'top', 'pod',
    *(('-n', namespace) if namespace else tuple()),
    *[arg for svc in services for arg in ('-l', f"app.kubernetes.io/name={svc}",)],
  ])



from kube_compose.cli import cli
from kube_compose import utils

@cli.group()
def get():
  ''' Access helm chart values
  '''
  pass

@get.command('all')
@utils.require_binaries(helm='helm')
@utils.require_kube_compose_release
def get_all(*, name, namespace, helm, **_):
  ''' download all information for a named release
  '''
  utils.run([*helm, 'get', 'all', *(('-n', namespace) if namespace else tuple()), name])

@get.command('manifest')
@utils.require_binaries(helm='helm')
@utils.require_kube_compose_release
def get_manifest(*, name, namespace, helm, **_):
  ''' download the manifest for a named release
  '''
  utils.run([*helm, 'get', 'manifest', *(('-n', namespace) if namespace else tuple()), name])

@get.command('metadata')
@utils.require_binaries(helm='helm')
@utils.require_kube_compose_release
def get_metadata(*, name, namespace, helm, **_):
  ''' This command fetches metadata for a given release
  '''
  utils.run([*helm, 'get', 'metadata', *(('-n', namespace) if namespace else tuple()), name])

@get.command('values')
@utils.require_binaries(helm='helm')
@utils.require_kube_compose_release
def get_values(*, name, namespace, helm, **_):
  ''' download the values file for a named release
  '''
  utils.run([*helm, 'get', 'values', *(('-n', namespace) if namespace else tuple()), name])



import yaml
import tempfile
import pathlib
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(helm='helm', diff='diff')
@utils.require_kube_compose_release
def diff(*, name, namespace, docker_compose_config, helm, diff, **_):
  ''' Compare the local docker-compose with the one deployed to kubernetes
  '''
  helm_release_yaml = yaml.safe_load(utils.check_output([helm, 'get', 'values', *(('-n', namespace,) if namespace else tuple()), name]))
  del helm_release_yaml['USER-SUPPLIED VALUES']
  #
  with tempfile.TemporaryDirectory() as tmpdir:
    tmpdir = pathlib.Path(tmpdir)
    # write as canonical yaml
    with (tmpdir / 'local.docker-compose.yaml').open('w') as fw:
      yaml.dump(docker_compose_config, fw, sort_keys=True)
    with (tmpdir / 'remote.docker-compose.yaml').open('w') as fw:
      yaml.dump(helm_release_yaml, fw, sort_keys=True)
      utils.run([
        *diff, '-Naur', str(tmpdir / 'remote.docker-compose.yaml'), str(tmpdir / 'local.docker-compose.yaml')
      ])



import click
import json
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@click.option('-n', '--namespace', help='The namespace to restrict ls to')
@utils.require_binaries(helm='helm')
def ls(*, namespace, helm, **_):
  ''' Like `docker-compose ls` but shows the docker-compose releases
  '''
  utils.echo_table_via_pager(
    [('NAME', 'NAMESPACE', 'REVISION', 'UPDATED', 'STATUS')] + [
    (release['name'], release['namespace'], release['revision'], release['updated'], release['status'])
    for release in json.loads(
      utils.check_output([
        *helm, 'list',
        *(('-n', namespace) if namespace else ('-A',)),
        '-o', 'json',
      ])
    )
    if release['chart'].startswith('kube-compose-')
  ])



import click
import itertools
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
@click.argument('service', type=str, required=False)
def start(service, *, namespace, kubectl, docker_compose_config, **_):
  ''' Like `docker-compose start` but effects the kubernetes deployed resources
  '''
  if service is not None:
    replicas = docker_compose_config.get('services', {}).get(service, {}).get('deploy', {}).get('replicas', 1) or 1
    utils.run([
      *kubectl, 'scale',
      *(('-n', namespace) if namespace else tuple()),
      f"--replicas={replicas}",
      f"deploy/{service}",
    ])
  else:
    service_replicas = []
    for svc in docker_compose_config.get('services', {}).keys():
      replicas = docker_compose_config['services'][svc].get('deploy', {}).get('replicas', 1)
      service_replicas.append((replicas, svc))
    for replicas, svcs in itertools.groupby(sorted(service_replicas), lambda key: key[0]):
      utils.run([
        *kubectl, 'scale',
        *(('-n', namespace) if namespace else tuple()),
        f"--replicas={replicas}",
        *[f"deploy/{svc}" for svc in svcs],
      ])



from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@utils.require_binaries(helm='helm')
@utils.require_kube_compose_release
def history(*, helm, name, namespace, **_):
  ''' Alias for `helm history [release]`
  '''
  utils.run([
    *helm, 'history',
    *(('-n', namespace) if namespace else tuple()),
     name,
  ])



import importlib
from kube_compose.cli import cli
from kube_compose import utils

for f in (utils.__dir__ / 'cli').rglob('[!_]*.py'):
  importlib.import_module('.'.join([
    utils.__dir__.stem, *f.relative_to(utils.__dir__).parts[:-1], f.stem
  ]))

if __name__ == '__main__':
  cli()



import click
from kube_compose.cli import cli
from kube_compose import utils

@cli.command()
@click.option('-v', '--volumes', type=bool, is_flag=True, help='Remove named volumes declared in the "volumes" section')
@utils.require_binaries(helm='helm')
@utils.require_kube_compose_release
def down(*, helm, name, namespace, volumes, **_):
  ''' Like `docker-compose down` but stops the kubernetes deployed resources
  '''
  utils.run([
    *helm, 'uninstall',
    *(('-n', namespace) if namespace else tuple()),
    name,
  ])
  if volumes:
    from kube_compose.cli.volume.rm import rm
    rm(volume=None)



import click
import pathlib
import functools

__dir__ = pathlib.Path(__file__).parent

helm_chart = str(__dir__ / 'charts' / 'kube-compose')

def assert_defined(value, or_else):
  assert value, or_else
  return value

def pathlib_coalesce(*paths):
  for path in map(pathlib.Path, paths):
    if path.exists(): return path
  if not paths: raise RuntimeError('No paths to check')
  else: raise FileNotFoundError(paths[0])

def set_group_config(fn):
  @click.pass_context
  def wrapper(ctx, *args, **kwargs):
    ctx.ensure_object(dict)
    ctx.obj.update(**kwargs)
    return ctx.invoke(fn, *args, **kwargs)
  return functools.update_wrapper(wrapper, fn)

def require_group_config(fn):
  @click.pass_context
  def wrapper(ctx: click.Context, *args, **kwargs):
    return ctx.invoke(fn, *args, **dict(ctx.obj, **kwargs))
  return functools.update_wrapper(wrapper, fn)

@click.pass_context
def run(ctx: click.Context, *args, input=None, **kwargs):
  ''' Similar to subprocess.run but
  1. forward Ctrl+C
  2. exit click application with exitcode of the process
  '''
  import signal
  import subprocess
  if input: kwargs['stdin'] = subprocess.PIPE
  p = subprocess.Popen(*args, **kwargs)
  if input:
    p.stdin.write(input)
    p.stdin.close()
  while p.returncode is None:
    try:
      p.wait()
    except KeyboardInterrupt:
      p.send_signal(signal.SIGINT)
    except:
      p.kill()
      p.wait()
      raise
  if p.returncode != 0:
    ctx.exit(p.returncode)

@click.pass_context
def check_output(ctx: click.Context, *args, **kwargs):
  ''' Similar to subprocess.check_output but exit click application with exitcode of the process on error
  '''
  import subprocess
  try:
    return subprocess.check_output(*args, **kwargs)
  except subprocess.CalledProcessError as e:
    ctx.exit(e.returncode)

def require_binaries(**required_binaries):
  ''' Ensure binaries are present, and inject the paths to them in kwargs
  Usage:
  @require_binaries(curl_bin='curl -s')
  def whats_my_ip(*, curl_bin):
    return check_output([*curl_bin, 'https://ifconfig.me']).decode()
  '''
  def decorator(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
      from shutil import which
      binaries = {
        arg: [assert_defined(which(bin), f"{bin} command not found"), *args]
        for arg, binary in required_binaries.items()
        for bin, *args in (binary if type(binary) == list else binary.split(' '),)
      }
      return fn(*args, **dict(binaries, **kwargs))
    return wrapper
  return decorator

def echo_table_via_pager(records):
  import click
  colSize = {}
  for record in records:
    for i, cell in enumerate(record):
      colSize[i] = max(colSize[i], len(cell)) if i in colSize else len(cell)
  click.echo_via_pager([
    ' '.join(('{:^%d}' % (colSize[i])).format(cell) for i, cell in enumerate(record)) + '\n'
    for record in records
  ])

@require_group_config
def locate_docker_compose_path(*, file, **_):
  docker_compose_path = pathlib.Path(file) if file else pathlib_coalesce('docker-compose.yaml', 'docker-compose.yml')
  return docker_compose_path

def require_kube_compose_release(fn):
  @functools.wraps(fn)
  @require_binaries(docker_compose='docker compose')
  def wrapper(*, docker_compose, **kwargs):
    import yaml
    docker_compose_config_raw = check_output([*docker_compose, '-f', str(locate_docker_compose_path()), 'config'])
    docker_compose_config = yaml.safe_load(docker_compose_config_raw)
    if 'x-kubernetes' not in docker_compose_config or not docker_compose_config.get('name'):
      raise click.ClickException('top-level `x-kubernetes` map with release `name` is required')
    release_config = docker_compose_config['x-kubernetes']
    name = release_config['name']
    namespace = release_config.get('namespace')
    return fn(**dict(kwargs,
      docker_compose=docker_compose,
      docker_compose_config_raw=docker_compose_config_raw,
      docker_compose_config=docker_compose_config,
      release_config=release_config,
      name=name,
      namespace=namespace,
    ))
  return wrapper



import json
import click
from kube_compose import utils
from kube_compose.cli.volume import volume

@volume.command('ls')
def _(**kwargs):
  ''' List kubernetes persistent volume claims
  '''
  ls(**kwargs)

@utils.require_binaries(kubectl='kubectl')
@utils.require_kube_compose_release
def ls(*, namespace, kubectl, **_):
  utils.run([
    *kubectl, 'get',
    *(('-n', namespace) if namespace else tuple()),
    'pvc',
  ])
